"""
Copyright 2021 cpyberry
https://github.com/cpyberry/pybitfield

cpyberry
email: cpyberry222@gmail.com
github: https://github.com/cpyberry
"""


import math
from enum import Enum, auto


class BitOrder(Enum):
	big = auto()
	little = auto()


class Bitfield:
	def __init__(self, number_of_element: int, bitfield=0):
		"""
		Args:
			number_of_element (int): fixed length bitfield size.
			bitfield (int, optional): initial value of bitfield. Defaults to 0.
		"""
		self.number_of_element = number_of_element
		self.bitfield = bitfield

	def set_bit(self, index: int) -> None:
		"""Set the bit at the location specified by index.

		The index starts from zero.

		Args:
			index (int): where you want to set the bit.
		"""
		self.bitfield |= self.to_bitfield(index)

	def remove_bit(self, index: int) -> None:
		"""Remove the bit at the location specified by index.

		The index starts from zero.

		Args:
			index (int): where you want to remove the bit.
		"""
		self.bitfield &= ~self.to_bitfield(index)

	def is_bit(self, index: int) -> bool:
		"""Check if the bit specified by index is set.

		Args:
			index (int): Where you want to find out if a bit is setting.

		Returns:
			bool: True if bit is set, False otherwise.
		"""
		return self.bitfield & self.to_bitfield(index) != 0

	def get_bit_list(self) -> list:
		"""Return the current bit state as a bool type list.

		Returns:
			list[bool]: True if bit is set, False otherwise.
		"""
		result = []
		for index in range(self.number_of_element):
			result.append(self.is_bit(index))
		return result

	def get_bitfield_bytes(self, bit_order=BitOrder.big) -> bytes:
		"""Convert current bitfield to bytes type.

		Returns:
			bytes: bitfield converted to byte type.
		"""
		byte_length = math.ceil(self.number_of_element / 8)

		if bit_order == BitOrder.big:
			return self.bitfield.to_bytes(byte_length, byteorder="big")
		else:
			bit_length = byte_length * 8
			little_endian_bitfield = self.swap_bitfield(bit_length)
			# It has already been converted to little endian, so return it as it is as big endian.
			return little_endian_bitfield.to_bytes(byte_length, byteorder="big")

	def swap_bitfield(self, length=None) -> int:
		"""Returns the swapped bitfield.

		If 0b10 is specified as an argument, the return value will be 0b01.

		Args:
			length (int, optional): bit length. If None, the current bitfield length is applied. Defaults to None.

		Returns:
			int: swapped bitfield.
		"""
		return self.swap_any_bitfield(self.bitfield, length)

	@classmethod
	def from_bytes(cls, data: bytes, length: int, bit_order=BitOrder.big):
		"""Create an instance of the Bitfield class from the bytes type.

		When the bytes is converted to bitfield, bits which exceed max bit length are removed.

		Args:
			data (bytes): bytes to convert to bitfield.
			length (int): fixed length size of bitfield.
			bit_order (BitOrder, optional): bit order. Defaults to BitOrder.big.

		Returns:
			Bitfield: an instance of the Bitfield class.
		"""
		bitfield = int.from_bytes(data, byteorder="big")

		if bit_order == BitOrder.little:
			bitfield = cls.swap_any_bitfield(bitfield)

		bitfield_bit_length = bitfield.bit_length()

		if bitfield_bit_length > length:
			fitted_bitfield = bitfield >> bitfield_bit_length - length
		else:
			fitted_bitfield = bitfield

		return Bitfield(length, fitted_bitfield)

	@staticmethod
	def to_bitfield(index: int) -> int:
		"""Return a bitfield with a bit at the position specified by index.

		The index starts from zero.

		Args:
			index (int): where you want to set the bit.

		Returns:
			int: the bitfield.
		"""
		return pow(2, index)

	@staticmethod
	def swap_any_bitfield(bitfield: int, length=None) -> int:
		"""Returns the swapped bitfield.

		If 0b10 is specified as an argument, the return value will be 0b01.

		Args:
			bitfield (int): bitfield you want to swap.
			length (int, optional): bit length. If None, the current bitfield length is applied. Defaults to None.

		Returns:
			int: swapped bitfield.
		"""
		if not length:
			length = len(str(bitfield))

		bitfield_str = format(bitfield, "b")
		swapped_bitfield_str = bitfield_str.zfill(length)[::-1]
		return int(swapped_bitfield_str, 2)



"""
Copyright 2021 cpyberry
https://github.com/cpyberry/pybitfield

cpyberry
email: cpyberry222@gmail.com
github: https://github.com/cpyberry
"""


__title__ = "pybitfield"
__description__ = "A library that makes it easy to handle bitfields with python"
__copyright__ = "Copyright 2021 cpyberry"
__url__ = "https://github.com/cpyberry/pybitfield"
__license__ = "Apache-2.0 License"
__version__ = "1.0.1"
__author__ = "cpyberry"
__author_email__ = "cpyberry222@gmail.com"
__python_requires__ = ">=3.6"



"""
Copyright 2021 cpyberry
https://github.com/cpyberry/pybitfield

cpyberry
email: cpyberry222@gmail.com
github: https://github.com/cpyberry
"""


import os
from setuptools import setup


NAME = "pybitfield"


def get_absolute_path(*paths):
	root_dir = os.path.abspath(os.path.dirname(__file__))
	path = os.path.join(root_dir, *paths)
	return path


info = {}
with open(get_absolute_path(NAME, "__info__.py")) as f:
	exec(f.read(), info)


with open(get_absolute_path("README.md"), encoding="utf-8") as f:
	long_description = f.read()


setup(
	name=info["__title__"],
	description=info["__description__"],
	long_description=long_description,
	long_description_content_type="text/markdown",
	url=info["__url__"],
	license=info["__license__"],
	version=info["__version__"],
	author=info["__author__"],
	author_email=info["__author_email__"],
	packages=[NAME],
	python_requires=info["__python_requires__"],
	zip_safe=False,
	include_package_data=True,
	keywords=["bitfield"],
	classifiers=[
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: Apache Software License",
		"Natural Language :: English",
		"Natural Language :: Japanese",
		"Programming Language :: Python",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"Programming Language :: Python :: Implementation",
		"Topic :: Software Development :: Libraries :: Python Modules"
	]
)



"""
Copyright 2021 cpyberry
https://github.com/cpyberry/pybitfield

cpyberry
email: cpyberry222@gmail.com
github: https://github.com/cpyberry
"""


from .bitfield import BitOrder, Bitfield

from .__info__ import (
	__title__,
	__description__,
	__copyright__,
	__url__,
	__license__,
	__version__,
	__author__,
	__author_email__,
	__python_requires__
)
